@@ -1,25 +1,30 @@
 """DB Hydra ORM"""
-import pymongo
-import pyodbc
-import pandas as pd
-import numpy as np
-import pymysql as MySQLdb
+import abc
+import ast
+import json
+import math
+import sys
+import threading
+from contextlib import contextmanager
 from sys import platform
+
+import numpy as np
+import pandas as pd
+import pymongo
+import pymysql as MySQLdb
+from google.cloud import bigquery
+from google.oauth2 import service_account
+from pydantic import BaseModel
+
+#! Disabled on macOS --> problematic import
+if sys.platform != "darwin":
+    import pyodbc
+
 if platform != "linux" and platform != "linux2":
     # linux
     import psycopg2 # disable dependency for server (Temporary)
 
 
-import math
-import json
-import ast
-
-from google.cloud import bigquery
-from google.oauth2 import service_account
-
-import abc
-from contextlib import contextmanager
-
 MONGO_OPERATOR_DICT = {"=": "$eq", ">": "$gt", ">=": "$gte", " IN ": "$in", "<": "$lt", "<=": "$lte", "<>": "$ne"}
 
 POSTGRES_TO_MYSQL_DATA_MAPPING = {
@@ -41,7 +46,7 @@ PYTHON_TO_MYSQL_DATA_MAPPING = {
     'int': "int",
     'float': "double",
     'str': "varchar(255)",
-    'bool': "tinyint",
+    'bool': "bit",
     'datetime': "datetime"
 }
 
@@ -101,7 +106,7 @@ def save_migration(function, *args, **kw):  # decorator
 
         migrator = instance.db1.migrator
         migrator.migration_list.append(migration_dict)
-        migrator.migration_list_to_json()
+        # migrator.migration_list_to_json()
         function(instance, *args, **kw)
 
     return (new_function)
@@ -109,12 +114,13 @@ def save_migration(function, *args, **kw):  # decorator
 
 # class DataMigrator:
 
+class DbHydraException(Exception):
+    """Raise for db_hydra specific exceptions such as 'pyodbc can't be used on macOS as it's not supported'."""
 
 
 class Migrator:
-    def __init__(self, db=None, db2=None):
+    def __init__(self, db=None):
         self.db = db
-        self.db2 = db2
         self.migration_number = 1
         self.migration_list = []
 
@@ -236,6 +242,27 @@ class Migrator:
 
 
 class AbstractDB(abc.ABC):
+    typing_to_python_mapping = {
+        'List': 'list',
+        'Dict': 'dict',
+        'Tuple': 'tuple',
+        'Set': 'set',
+        'Union': 'object',
+        'Optional': 'object',
+        # 'FrozenSet': frozenset,
+        # 'Deque': list,
+        # 'Any': object,
+        #
+        # 'Callable': object,
+        # 'Type': type,
+        # 'TypeVar': object,
+        # 'Generic': object,
+        # 'Sequence': list,
+        # 'Iterable': list,
+        # 'Mapping': dict,
+        # 'AbstractSet': set,
+    }
+    python_database_type_mapping = {}
     def __init__(self, config_file="config.ini", db_details=None):
         if db_details is None:
             db_details = read_connection_details(config_file)
@@ -259,6 +286,8 @@ class AbstractDB(abc.ABC):
         else:
             self.DB_DRIVER = "ODBC Driver 13 for SQL Server"
 
+        self.lock = threading.Lock()
+
         self.connect_to_db()
 
     @abc.abstractmethod
@@ -318,16 +347,28 @@ class AbstractDB(abc.ABC):
 
             yield table
 
-    def execute(self, query):
-        self.cursor.execute(query)
-        self.cursor.commit()
+    def execute(self, query, is_autocommitting=True):
+        result=self.cursor.execute(query)
+        if is_autocommitting:
+            self.cursor.commit()
+        return(result)
 
     def close_connection(self):
         self.connection.close()
         print("DB connection closed")
 
-    def initialize_migrator(self, other_db=None):
-        self.migrator = Migrator(self, other_db)
+    def initialize_migrator(self):
+        self.migrator = Migrator(self)
+
+    def _convert_column_type_dict_from_python(self, column_type_dict):
+        """
+        First apply mapping from python typing module to standard python.
+        Then apply mapping from python to database types
+        """
+        typing_python_mapping = {column_name: self.typing_to_python_mapping.get(column_type,column_type) for column_name, column_type in column_type_dict.items()}
+        return {column_name: self.python_database_type_mapping[column_type] for column_name, column_type in typing_python_mapping.items()}
+
+
 
 
 class AbstractDBPostgres(AbstractDB):
@@ -350,6 +391,8 @@ class AbstractDBMongo(AbstractDB):
 # AWFUL NAME, SHOULD BE RENAMED WITH MAJOR RELEASE (This connects basic SQL i suppose)
 class db(AbstractDB):
     def connect_remotely(self):
+        if sys.platform == "darwin":
+            raise DbHydraException("pyodbc library (MSSQL) not supported on macOS.")
 
         self.connection = pyodbc.connect(
             r'DRIVER={' + self.DB_DRIVER + '};'
@@ -364,6 +407,9 @@ class db(AbstractDB):
         print("DB connection established")
 
     def connect_locally(self):
+        if sys.platform == "darwin":
+            raise DbHydraException("pyodbc library (MSSQL) not supported on macOS.")
+        
         self.connection = pyodbc.connect(
             r'DRIVER={' + self.DB_DRIVER + '};'
                                            r'SERVER=' + self.DB_SERVER + ';'
@@ -414,12 +460,29 @@ class db(AbstractDB):
 
 
 class Mysqldb(AbstractDB):
+
+    python_database_type_mapping = PYTHON_TO_MYSQL_DATA_MAPPING = \
+    {
+    'int': "int",
+    'float': "double",
+    'str': "varchar(255)",
+    'list': "varchar(255)",
+    'dict': "varchar(255)",
+    'bool': "tinyint",
+    'datetime': "datetime"
+    }
+
+    def connect_to_db(self):
+        self.connection = MySQLdb.connect(host=self.DB_SERVER, port=self.DB_PORT, user=self.DB_USERNAME, password=self.DB_PASSWORD, database=self.DB_DATABASE)
+        self.cursor = self.connection.cursor()
+
+    # NOT USED, BUT FORCED BY ABSTRACT CLASS
     def connect_locally(self):
         self.connection = MySQLdb.connect(host=self.DB_SERVER, user=self.DB_USERNAME, password=self.DB_PASSWORD,
                                           database=self.DB_DATABASE)
         self.cursor = self.connection.cursor()
         print("DB connection established")
-
+    # NOT USED, BUT FORCED BY ABSTRACT CLASS
     def connect_remotely(self):
         if self.DB_PORT is not None:
             self.connection = MySQLdb.connect(host=self.DB_SERVER, port=self.DB_PORT, user=self.DB_USERNAME,
@@ -441,10 +504,12 @@ class Mysqldb(AbstractDB):
 
         self.connection.commit()
 
-    def execute(self, query):
-
-        self.cursor.execute(query)
-        self.connection.commit()
+    def execute(self, query, is_autocommitting=True):
+        result=self.cursor.execute(query)
+        if is_autocommitting:
+            self.connection.commit()
+        return result
+    
 
     def get_all_tables(self):
         sysobjects_table = Table(self, "information_schema.tables", ["TABLE_NAME"], ["nvarchar(100)"])
@@ -480,9 +545,11 @@ class PostgresDb(AbstractDBPostgres):
         self.connection.autocommit = True
         self.cursor = self.connection.cursor()
 
-    def execute(self, query):
-        self.cursor.execute(query)
-        self.connection.commit()
+    def execute(self, query, is_autocommitting=True):
+        result=self.cursor.execute(query)
+        if is_autocommitting:
+            self.connection.commit()
+        return result
 
         # return  [''.join(i) for i in self.cursor.fetchall()]
 
@@ -502,11 +569,12 @@ class PostgresDb(AbstractDBPostgres):
 
 
 class BigQueryDb(AbstractDB):
-    def __init__(self, credentials_path, project_id, dataset_name):
-        # super().__init__(None,None)
-        self.credentials_path = credentials_path
-        self.project_id = project_id
-        self.dataset = dataset_name
+    def __init__(self, db_details):
+        self.credentials_path = db_details["DB_SERVER"]
+        self.dataset = db_details["DB_DATABASE"]
+
+
+        self.locally = False
 
         self.credentials = service_account.Credentials.from_service_account_file(self.credentials_path, scopes=[
             "https://www.googleapis.com/auth/cloud-platform"], )
@@ -518,6 +586,7 @@ class BigQueryDb(AbstractDB):
     def connect_locally(self):
         raise Exception("Cannot connect locally to Big Query")
 
+
     def close_connection(self):
         self.client.close()
 
@@ -548,27 +617,9 @@ class BigQueryDb(AbstractDB):
 
 
     def execute(self, query):
-        # query_job = self.client.query(
-        #     """
-        #     SELECT
-        #       CONCAT(
-        #         'https://stackoverflow.com/questions/',
-        #         CAST(id as STRING)) as url,
-        #       view_count
-        #     FROM `bigquery-public-data.stackoverflow.posts_questions`
-        #     WHERE tags like '%google-bigquery%'
-        #     ORDER BY view_count DESC
-        #     LIMIT 10"""
-        # )
-
         query_job = self.client.query(query)
-
         results = query_job.result()  # Waits for job to complete.
-
-        for row in results:
-            print(row)
-
-            # print("{} : {} views".format(row.id, row.link,row.title))
+        return results
 
 
 class MongoDb(AbstractDBMongo):
@@ -593,10 +644,12 @@ class MongoDb(AbstractDBMongo):
         print(self.database)
         print("DB connection established")
 
-    def execute(self, query):
-        self.cursor.execute(query)
-        self.connection.commit()
-
+    def execute(self, query, is_autocommitting=True):
+        result=self.cursor.execute(query)
+        if is_autocommitting:
+            self.connection.commit()
+        return(result)
+    
     def get_all_tables(self):
         return self.database.list_collection_names()
 
@@ -629,7 +682,6 @@ class AbstractSelectable:
         self.columns = columns
 
     def select(self, query):
-
         """given SELECT query returns Python list"""
         """Columns give the number of selected columns"""
         print(query)
@@ -666,7 +718,6 @@ class AbstractSelectable:
         list1 = self.select(f"SELECT {all_cols_query} FROM " + self.name)
         return (list1)
 
-
     def select_to_df(self):
         rows = self.select_all()
         table_columns = self.columns
@@ -707,12 +758,15 @@ class Joinable(Selectable):
     pass
 
 
-class AbstractTable(AbstractJoinable):
+class AbstractTable(AbstractJoinable, abc.ABC):
     def __init__(self, db1, name, columns=None, types=None):
         super().__init__(db1, name, columns)
         self.types = types
 
-    # @save_migration
+    # Temporary disabled, please make sure this is implemented where needed, don't introduce breaking changes please
+    # @abc.abstractmethod
+    def init_from_column_type_dict(db1, name, column_type_dict):
+        pass
 
     def drop(self):
         query = "DROP TABLE " + self.name
@@ -725,7 +779,7 @@ class AbstractTable(AbstractJoinable):
         else:
             query = "UPDATE " + self.name + " SET " + variable_assign + " WHERE " + where
         print(query)
-        self.db1.execute(query)
+        return self.db1.execute(query)
 
     def _adjust_df(self, df: pd.DataFrame, debug_mode=False) -> pd.DataFrame:
         """
@@ -760,6 +814,7 @@ class AbstractTable(AbstractJoinable):
 
         return df_copy
 
+
     def insert_from_df(self, df, batch=1, try_mode=False, debug_mode=False, adjust_df=False, insert_id=False):
 
         if adjust_df:
@@ -767,8 +822,22 @@ class AbstractTable(AbstractJoinable):
 
         if insert_id:
             assert len(df.columns) == len(self.columns)
+            assert set(df.columns) == set(self.columns)
+            
+            inserted_columns=self.columns
+            
         else:
             assert len(df.columns) + 1 == len(self.columns) # +1 because of id column
+            
+            inserted_columns=list(dict.fromkeys(self.columns)) #DEDUPLICATION preserving order -> better than inserted_columns = set(self.columns) 
+            id_index=inserted_columns.index("id")
+            inserted_columns.pop(id_index)
+            print(inserted_columns,df.columns)
+            
+            assert set(df.columns) == set(inserted_columns) #elements are matchin
+            #df = df[inserted_columns]
+        df = df[inserted_columns]
+
         pd_nullable_dtypes = {pd.Int8Dtype(), pd.Int16Dtype(), pd.Int32Dtype(), pd.Int64Dtype(),
                               pd.UInt8Dtype(), pd.UInt16Dtype(), pd.UInt32Dtype(), pd.UInt64Dtype(),
                               pd.Float32Dtype(), pd.Float64Dtype()}
@@ -783,12 +852,25 @@ class AbstractTable(AbstractJoinable):
         for column in list(df.columns):
             df.loc[pd.isna(df[column]), column] = "NULL"
 
+        # rows = df.values.tolist()
+        # for i, row in enumerate(rows):
+        #     for j, record in enumerate(row):
+        #         if type(record) == str:
+        #             rows[i][j] = "'" + record + "'"
+        #print(rows)
         rows = df.values.tolist()
-        for i, row in enumerate(rows):
-            for j, record in enumerate(row):
-                if type(record) == str:
-                    rows[i][j] = "'" + record + "'"
-        self.insert(rows, batch=batch, try_mode=try_mode, debug_mode=False, insert_id=insert_id)
+        return self.insert(rows, batch=batch, try_mode=try_mode, debug_mode=False, insert_id=insert_id)
+
+    #TODO: need to solve inserting in different column_order
+    #check df column names, permute if needed
+    def insert_from_column_value_dict(self, dict, insert_id=False):
+        df = pd.DataFrame(dict, index=[0])
+        return self.insert_from_df(df, insert_id=insert_id)
+
+    def insert_from_column_value_dict_list(self, list, insert_id=False):
+        df = pd.DataFrame(list)
+        self.insert_from_df(df, insert_id=insert_id)
+
 
     def delete(self, where=None):
 
@@ -796,8 +878,7 @@ class AbstractTable(AbstractJoinable):
             query = "DELETE FROM " + self.name
         else:
             query = "DELETE FROM " + self.name + " WHERE " + where
-        print(query)
-        self.db1.execute(query)
+        return self.db1.execute(query)
 
 
 class PostgresTable(AbstractTable):
@@ -814,9 +895,6 @@ class PostgresTable(AbstractTable):
         self.columns = columns
 
     def initialize_types(self):
-        # information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
-        # query = f"SELECT DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
-        # types = information_schema_table.select(query)
         self.types = self.get_all_types()
 
     def get_all_columns(self):
@@ -836,15 +914,15 @@ class PostgresTable(AbstractTable):
         information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
         query = "SELECT DATA_TYPE,character_maximum_length FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_NAME  = '" + self.name + "'"
         types = information_schema_table.select(query)
-        types = [x[0] for x in types]
-        types = [x.lower() for x in types]
-        lengths = [x[1] for x in types]
+        data_types = [x[0].lower() for x in types]
+        date_lengths = [x[1] for x in types]
+
+        mysql_types = list(map(lambda x: POSTGRES_TO_MYSQL_DATA_MAPPING.get(x, x), data_types))
 
-        mysql_types = list(map(lambda x: POSTGRES_TO_MYSQL_DATA_MAPPING.get(x, x), types))
 
         for i in range(len(mysql_types)):
-            if lengths[i] is not None:
-                mysql_types[i] = mysql_types[i] + f"({lengths[i]})"
+            if date_lengths[i] is not None:
+                mysql_types[i] = mysql_types[i] + f"({date_lengths[i]})"
 
         return (mysql_types)
 
@@ -953,10 +1031,10 @@ class PostgresTable(AbstractTable):
                     print(query)
 
                 if not try_mode:
-                    self.db1.execute(query)
+                    return self.db1.execute(query)
                 else:
                     try:
-                        self.db1.execute(query)
+                        return self.db1.execute(query)
                     except Exception as e:
 
                         print("Query", query, "Could not be inserted:", e)
@@ -1002,10 +1080,10 @@ class PostgresTable(AbstractTable):
         except Exception as e:
             print("Cant add column to table.")
 
-class BigQueryTable():
-    def __init__(self, db, name, columns=None, types=None):
+class BigQueryTable(AbstractSelectable):
+    def __init__(self, db1, name, columns=None, types=None):
         self.name = name
-        self.db = db
+        self.db1 = db1
         self.columns = columns
         self.types = types
 
@@ -1017,12 +1095,23 @@ class BigQueryTable():
         return (cls(db1, name, columns, types))
 
     def get_all_columns_and_types(self):
-        results = self.db.client.get_table(self.name)
+        results = self.db1.client.get_table(f"{self.db1.credentials.project_id}.{self.db1.dataset}.{self.name}")
         column_names = [x.name for x in results.schema ]
         column_types = [x.field_type for x in results.schema]
         return column_names, column_types
 
 
+    def select(self, query):
+
+        """given SELECT query returns Python list"""
+        """Columns give the number of selected columns"""
+        print(query)
+        # rows =  self.db1.client.query(query).result()
+        rows = self.db1.execute(query)
+        return rows
+
+
+
 
 class MongoTable():
     def __init__(self, db, name, columns=[], types=[]):
@@ -1310,6 +1399,13 @@ class MysqlTable(MysqlSelectable, AbstractTable):
         super().__init__(db1, name, columns)
         self.types = types
 
+    @classmethod
+    def init_from_column_type_dict(cls, db1, name, column_type_dict):
+        column_converted_type_dict = db1._convert_column_type_dict_from_python(column_type_dict)
+        columns = list(column_converted_type_dict.keys())
+        types = list(column_converted_type_dict.values())
+        return cls(db1, name, columns, types)
+
     def initialize_columns(self):
         information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
         query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND  TABLE_NAME  = '" + self.name + "';"
@@ -1320,9 +1416,6 @@ class MysqlTable(MysqlSelectable, AbstractTable):
         pass
 
     def initialize_types(self):
-        # information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
-        # query = f"SELECT COLUMN_TYPE FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND TABLE_NAME  = '" + self.name + "'"
-        # types = information_schema_table.select(query)
         self.types = self.get_all_types()
 
     def get_all_columns(self):
@@ -1334,17 +1427,46 @@ class MysqlTable(MysqlSelectable, AbstractTable):
 
     def get_all_types(self):
 
+        data_types, data_lengths = self.get_data_types_and_character_lengths()
+        for i in range(len(data_types)):
+            if data_lengths[i] is not None:
+                data_types[i] = data_types[i] + f"({data_lengths[i]})"
+        return (data_types)
+
+
+    """
+        Returns a list of data types, where each element represents the category of the data ('varchar', 'int', etc.). 
+        If a data type has an associated length, the length value will be included in a corresponding element of the
+        data_lengths list, otherwise the element will have a None value. For example, 'varchar(255)' would return
+        'varchar' in the data_types list and 255 in the data_lengths list.
+    """
+    def get_data_types_and_character_lengths(self):
         information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS', ['DATA_TYPE'], ['nvarchar(50)'])
         query = f"SELECT DATA_TYPE,character_maximum_length FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA = '{self.db1.DB_DATABASE}' AND TABLE_NAME  = '" + self.name + "'"
         types = information_schema_table.select(query)
-        only_types = [x[0] for x in types]
-        lengths = [x[1] for x in types]
+        data_types = [x[0] for x in types]
+        data_lengths = [x[1] for x in types]
 
-        for i in range(len(only_types)):
-            if lengths[i] is not None:
-                only_types[i] = only_types[i] + f"({lengths[i]})"
-        return (only_types)
+        return data_types,data_lengths
 
+    def get_converted_python_types(self):
+        SQL_TO_PYTHON = {v: k for k, v in PYTHON_TO_MYSQL_DATA_MAPPING.items()}
+        python_types = []
+        for type in self.types:
+            if "varchar" in type:
+                python_types.append("str")
+            elif type in SQL_TO_PYTHON:
+                python_types.append(SQL_TO_PYTHON[type])
+            else:
+                raise Exception("Unsupported type")
+
+        return python_types
+
+    def get_nullable_columns(self):
+        information_schema_table = Table(self.db1, 'INFORMATION_SCHEMA.COLUMNS')
+        query = f"SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS where TABLE_SCHEMA = '{self.db1.DB_DATABASE}' and TABLE_NAME = '{self.name}' and IS_NULLABLE = 'YES'"
+        nullable_columns = information_schema_table.select(query)
+        return (nullable_columns)
 
 
     @classmethod
@@ -1373,6 +1495,15 @@ class MysqlTable(MysqlSelectable, AbstractTable):
 
         return last_id[0][0]
 
+    def get_num_of_records(self):
+        """
+        Returns the number of records in table
+        """
+
+        num_of_records = self.select(f"SELECT COUNT(*) FROM {self.name};")
+
+        return num_of_records[0][0]
+
     def drop(self):
         query = "DROP TABLE " + self.name + ";"
         print(query)
@@ -1383,12 +1514,12 @@ class MysqlTable(MysqlSelectable, AbstractTable):
         assert len(self.columns) == len(self.types)
         assert self.columns[0] == "id"
         assert self.types[0].lower() == "int"
-        query = "CREATE TABLE " + self.name + "(id INT UNSIGNED AUTO_INCREMENT PRIMARY KEY,"
-        for i in range(1, len(self.columns)):
-            query += self.columns[i] + " " + self.types[i] + ","
 
-        query = query[:-1]
-        query += ")"
+        column_type_pairs = list(zip(self.columns, self.types))[1:]
+        fields = ", ".join(
+            [f"{column} {type_.upper()}" for column, type_ in column_type_pairs]
+        )
+        query = f"CREATE TABLE {self.name} (id INT UNSIGNED AUTO_INCREMENT PRIMARY KEY, {fields})"
 
         print(query)
         try:
@@ -1457,10 +1588,10 @@ class MysqlTable(MysqlSelectable, AbstractTable):
                     print(query)
 
                 if not try_mode:
-                    self.db1.execute(query)
+                    return self.db1.execute(query)
                 else:
                     try:
-                        self.db1.execute(query)
+                        return self.db1.execute(query)
                     except Exception as e:
 
                         print("Query", query, "Could not be inserted:", e)
@@ -1662,4 +1793,24 @@ def df_to_dict(df, column1, column2):
 
 def dict_to_df(dictionary, column1, column2):
     df = pd.DataFrame(list(dictionary.items()), columns=[column1, column2])
-    return (df)
\ No newline at end of file
+    return (df)
+
+
+class AbstractModel(abc.ABC, BaseModel):
+    @classmethod
+    def generate_dbhydra_table(cls, table_class: AbstractTable, db1, name):
+        column_type_dict = create_table_structure_dict(cls)
+        dbhydra_table = table_class.init_from_column_type_dict(db1, name, column_type_dict)
+        return dbhydra_table
+
+ 
+
+def create_table_structure_dict(api_class_instance):
+    """
+    Accepts instance of API data class (e.g. APIDatabase) and converts it to dictionary {attribute_name: attribute_type}
+    """
+    table_structure_dict = {"id": "int"}
+    table_structure_dict = {**table_structure_dict,
+                            **{attribute_name: attribute_type.__name__ for attribute_name, attribute_type in api_class_instance.__annotations__.items()}}
+
+    return table_structure_dict
